// --------------------------------------------------------------------------
// External files inclusions.
// --------------------------------------------------------------------------
#include <pic.h>
#include "delay.h"



// --------------------------------------------------------------------------
// Required variables (declared as global because of ASM code). Not exported.
// --------------------------------------------------------------------------
unsigned long int counter;                                      // Denominator and reminder of division.
unsigned char status_tmp;                                       // Used to temporary store the STATUS register.



// --------------------------------------------------------------------------
// This procedure waits a given amount of instruction cycles. This time is
// unfortunately dependent on the clock frequency and it's not interrupt
// compatible.
// This procedure takes exactly 25 instruction cycles per value of "count",
// plus 32 to 35 instruction cycles for calling and returning (this fixed
// amount also depends on compiler options).
// The time required for calling and returning is unfortunately dependent on
// the C compiler and is probably neglectable but if this is not the case the
// table below will help:
//
// Way of calling delay()                   Total instruction cycles
// ------------------------------------     ------------------------
// delay(0);                                32
// delay(1); ... delay(255);                33 + (25 * delay)
// delay(256); ... delay(65535);            34 + (25 * delay)
// delay(65536); ... delay(4294967295);     35 + (25 * delay)
// delay(char_variable);                    32 + (25 * delay)
// delay(int_variable);                     33 + (25 * delay)
// delay(long_int_variable);                35 + (25 * long_int_variable)
//
// The NOP instruction can be replaced with a CLRWDT instruction to make
// this procedure watchdog compatible.
// --------------------------------------------------------------------------
void delay(unsigned long int count)
{
    counter = count;

#asm
; Make sure we are in bank 0 in order to correctly access variables.
                swapf   0x03,w                                  ; Save STATUS into W.
                bcf     0x03,5                                  ; Select bank 0 (this will work for any PIC16xxx).
                bcf     0x03,6
                movwf   (_status_tmp & 0x7F)                    ; And save into _status_tmp

; The following code will check if the 32 bit unsigned variable "counter" is
; zero or not. This code has a fixed length (it takes the same amount of
; instructions, no matter the value of "counter".
loop_again:     movlw   0x00                                    ; Start by putting 0 in register W.
                iorwf   ((_counter & 0x7F)+0),w                 ; Perform an inclusive OR function with
                iorwf   ((_counter & 0x7F)+1),w                 ; W and all 4 bytes of "counter"; result
                iorwf   ((_counter & 0x7F)+2),w                 ; will be 0 only if "counter" is 0.
                iorwf   ((_counter & 0x7F)+3),w
                btfsc   0x03,2                                  ; Check zero flag.
                goto    loop_done                               ; Counter is 0: exit delay loop.

; The following code will decrement the 32 bit unsigned variable "counter"
; by 1. This code has a fixed length (it takes the same amount of
; instructions, no matter the value of "counter".
                movlw   0x01                                    ; Decrement least significant byte.
                subwf   ((_counter & 0x7F)+0),f

                clrw                                            ; Prepare to substract 0 from next byte.
                btfss   0x03,0                                  ; Check carry flag.
                movlw   0x01                                    ; If no carry was set, the subtract 1 from next byte.
                subwf   ((_counter & 0x7F)+1),f                 ; Perform subtraction.

                clrw                                            ; Prepare to substract 0 from next byte.
                btfss   0x03,0                                  ; Check carry flag.
                movlw   0x01                                    ; If no carry was set, the subtract 1 from next byte.
                subwf   ((_counter & 0x7F)+2),f                 ; Perform subtraction.

                clrw                                            ; Prepare to substract 0 from most significant byte.
                btfss   0x03,0                                  ; Check carry flag.
                movlw   0x01                                    ; If no carry was set, the subtract 1 from most significant byte.
                subwf   ((_counter & 0x7F)+3),f                 ; Perform subtraction.

                nop                                             ; Just to make the loop 25 cycles long instead of 23.
                nop                                             ; One NOP be replaced with CLRWDT to make this procedure watchdog compliant.
                goto    loop_again                              ; Loop again

; Make sure we set back the bank that was in use before.
loop_done:      swapf   (_status_tmp & 0x7F),w                  ; Restore STATUS register.
                movwf   0x03
#endasm

}


