/*===========================================================================



Digital SWR meter
=================



Overview:
=========

CPU         :   PIC16F876A
Clock       :   XTAL 18.432MHz HS oscillator
Compiler    :   Hitech PICC
Version     :   1.00
Date        :   31.10.2008
Author      :   Iacopo Giangrandi HB9DUL



I/O connections:
================

See io.h for description.



Description:
============

...



============================================================================= */




// --------------------------------------------------------------------------
// External files inclusions.
// --------------------------------------------------------------------------
#include <pic.h>                                                // Use a PIC16F877A.
#include <math.h>
#include "io.h"                                                 // I/O lines definitions.
#include "delay.h"                                              // Delay procedure.
#include "lcd.h"                                                // LCD display handling routines.
#include "adc.h"                                                // AD converter.



// --------------------------------------------------------------------------
// Configuration bits definitions.
// --------------------------------------------------------------------------
__CONFIG(HS & WDTDIS & PWRTEN & BOREN & LVPDIS & UNPROTECT & DEBUGDIS);
// Use external XTAL 18.432 MHz HS oscillator, do not use watchdog, enable power
// up timer, enable brown out detection, disable low voltage programming, do not
// use data protection and disable debugging.



// --------------------------------------------------------------------------
// VFO definitions, constants and global variables.
// --------------------------------------------------------------------------
#define WELCOME_STRING_1    " Digital VSWR meter "
#define WELCOME_STRING_2    "v1.0 08.11.08 HB9DUL"
#define WELOCME_TIME        184320								// 1 second.



// --------------------------------------------------------------------------
// This procedure converts the specified power into a 4 character ASCII
// string and prints it on the LCD screen. Power must be positive and smaller
// than 999. Leading zeros are removed. Decimal point is insered according
// to DP_0 and DP_1. The unit ("W") is added after the number.
// --------------------------------------------------------------------------
void write_power(signed int a)
{
    unsigned char ascii_num[5];                                 // Four characters string used to display the number in formatted ASCII decimal format.
    signed char i;                                              // Just a counter.

    if ((a > 999) || (a < 0))                                   // Make sure power is not too big.
    {
        LCD_write_string("OVR ");                               // Write result on LCD screen.
        LCD_write_string(" W");
        return;
    }

    if (DP_1 == 1)
    {
        if (DP_0 == 1)
        {
            for (i=3; i>=0; i--)                                // Convert into an ASCII string.
            {
                if (i == 1)                                     // Insert decimal point ("x.xx").
                    ascii_num[i] = '.';
                else
                {
                    ascii_num[i] = (a % 10) + '0';              // Get ASCII digit.
                    a /= 10;
                }
            }
        }
        else
        {
            for (i=3; i>=0; i--)                                // Convert into an ASCII string.
            {
                if (i == 2)                                     // Insert decimal point ("xx.x").
                    ascii_num[i] = '.';
                else
                {
                    ascii_num[i] = (a % 10) + '0';              // Get ASCII digit.
                    a /= 10;
                }
            }
            if (ascii_num[0] == '0')                            // Remove leading 0.
                ascii_num[0] = ' ';
        }
    }
    else
    {
        if (DP_0 == 1)
        {
            for (i=3; i>=0; i--)                                // Convert into an ASCII string.
            {
                if (i == 0)                                     // Insert space (no decimal point) (" xxx").
                    ascii_num[i] = ' ';
                else
                {
                    ascii_num[i] = (a % 10) + '0';              // Get ASCII digit.
                    a /= 10;
                }
            }
            for (i=1; i<3; i++)                                 // Remove leading zeros.
                if (ascii_num[i] == '0')                        // Replace 0 with spaces.
                    ascii_num[i] = ' ';
                else
                    break;                                      // Until the first non-zero digit is found.
        }
        else
        {
            for (i=3; i>=0; i--)                                // Convert into an ASCII string.
            {
                if (i == 3)                                     // Insert zero at the end (no decimal point) ("xxx0").
                    ascii_num[i] = '0';
                else
                {
                    ascii_num[i] = (a % 10) + '0';              // Get ASCII digit.
                    a /= 10;
                }
            }
            for (i=0; i<3; i++)                                 // Remove leading zeros.
                if (ascii_num[i] == '0')                        // Replace 0 with spaces.
                    ascii_num[i] = ' ';
                else
                    break;                                      // Until the first non-zero digit is found.
        }
    }    
    ascii_num[4] = 0x00;                                        // Null terminate the string.
    LCD_write_string(ascii_num);                                // Write result on LCD screen.
    LCD_write_string(" W");
}



// --------------------------------------------------------------------------
// This procedure converts the specified SWR into a 4 character ASCII
// string and prints it on the LCD screen divided by 10. SWR must be
// positive and smaller than 999. Leading zeros are removed. Decimal point
// is insered to obtain xxx.x.  The unit (":1") is added after the number.
// --------------------------------------------------------------------------
void write_swr(signed int a)
{
    unsigned char ascii_num[5];                                 // Four characters string used to display the number in formatted ASCII decimal format.
    signed char i;                                              // Just a counter.

    if ((a > 1000) || (a < 0))                                  // Make sure SWR is not infinite. 
    {
        LCD_write_string("----");
		LCD_write_string(":1");
        return;
    }

    if (a > 999)                                                // Make sure SWR is smaller than or equal to 9999.
        a = 999;

    for (i=3; i>=0; i--)                                        // Convert into an ASCII string.
    {
        if (i == 2)                                             // Insert decimal point ("xx.x").
            ascii_num[i] = '.';
        else
        {
            ascii_num[i] = (a % 10) + '0';                      // Get ASCII digit.
            a /= 10;
        }
    }

    if (ascii_num[0] == '0')                                    // Remove leading 0.
        ascii_num[0] = ' ';

    ascii_num[4] = 0x00;                                        // Null terminate the string.

    LCD_write_string(ascii_num);                                // Write result on LCD screen.
    LCD_write_string(":1");
}



// --------------------------------------------------------------------------
// This procedure converts the specified integer value into a 3 character
// ASCII string and prints it on the LCD screen. The value must be between
// -99 and +999. Leading zeros are removed and a minus sign is diplayed if
// needed. The unit (" dB") is added after the number.
// --------------------------------------------------------------------------
void write_db(signed int a)
{
    unsigned char ascii_num[4];                                 // Three characters string used to display the number in formatted ASCII decimal format.
    signed char i;                                              // Just a counter.
    unsigned char neg;

    if ((a < -99) || (a > 999))                                 // Make sure range is valid.
    {
        LCD_write_string("---");
        LCD_write_string(" dB");
        return;
    }

    neg = 0;
    if (a < 0)
    {
        neg = 1;
        a = -a;
    }

    for (i=2; i>=0; i--)                                        // Convert into an ASCII string.
    {
        ascii_num[i] = (a % 10) + '0';                          // Get ASCII digit.
        a /= 10;
    }

    for (i=0; i<2; i++)                                         // Remove leading zeros.
        if (ascii_num[i] == '0')                                // Replace 0 with spaces.
            ascii_num[i] = ' ';
        else
            break;                                              // Until the first non-zero digit is found.

    if (neg == 1)                                               // Place minus sign, if needed, right before the first digit.
    {
        if (ascii_num[1] == ' ')
            ascii_num[1] = '-';
        else
            ascii_num[0] = '-';
    }

    ascii_num[3] = 0x00;                                        // Null terminate the string.
    LCD_write_string(ascii_num);                                // Write result on LCD screen.
    LCD_write_string(" dB");
}



// --------------------------------------------------------------------------
// This procedure converts the specified SWR into a 10 character ASCII
// string containing a bar graph and prints it on the LCD screen.
// SWR is divided by 10.
// --------------------------------------------------------------------------
void write_swr_bar(signed int a)
{
    unsigned char ascii_num[11];                                // Ten characters string used to display the number in formatted ASCII decimal format.
    signed char i,j;                                            // Just a counter and a temporary value.

    if (a < 0)                                                  // If SWR is infinite (represented as -1), make sure the bar is at its full length.
        a = 30000;

    if (a < 11)                                                 // Use a non-linear scale, first brick at 1.1:1.
        j = 0;
    else if (a < 12)                                            // Second brick at 1.2:1.
        j = 1;
    else if (a < 15)
        j = 2;
    else if (a < 18)
        j = 3;
    else if (a < 20)
        j = 4;
    else if (a < 25)
        j = 5;
    else if (a < 30)
        j = 6;
    else if (a < 40)
        j = 7;
    else if (a < 60)
        j = 8;
    else if (a < 80)
        j = 9;
    else
        j = 10;

    for (i=0; i<10; i++)                                        // Draw the bar-graph.
    {
        if (i < j)
            ascii_num[9-i] = 0xFF;
        else
            ascii_num[9-i] = ' ';
    }

    ascii_num[10] = 0x00;                                       // Null terminate the string.
    LCD_write_string(ascii_num);                                // Write result on LCD screen.
}



// --------------------------------------------------------------------------
// This is the main program.
// --------------------------------------------------------------------------
void main(void)
{



    // Variables.
    
    signed long int pwr_fwd;
    signed long int pwr_rev;
    signed int swr;
    double a;


    
    // Initialization.

    IO_init();                                                  // Initialize I/O lines.
    LCD_init();                                                 // Initialize LCD display.



    // Welcome screen.

    LCD_LIGHT = LIGHT_ON;                                       // Switch LCD light on.
    LCD_clear();
    LCD_home();
    LCD_line_1();                                               // Display welcome screen.
    LCD_write_string(WELCOME_STRING_1);
    LCD_line_2();
    LCD_write_string(WELCOME_STRING_2);

    delay(WELOCME_TIME);                                        // Wait 1s



    // Main loop.

    while (1)                                                   // Endless loop tha polls PORTB waiting for user input.
    {
        pwr_fwd = adc_read(PWR_FWD);                            // Read volatges.
        pwr_rev = adc_read(PWR_REV);
        pwr_fwd *= pwr_fwd;                                     // Square to get power.
        pwr_rev *= pwr_rev;
        if ((SCALE_0 == 0) && (SCALE_1 == 1))
        {
            pwr_fwd *= 2;                                       // Full scale at 2W, 20W, 200W or 2kW.
            pwr_rev *= 2;
        }   
        else if ((SCALE_0 == 1) && (SCALE_1 == 0))
        {
            pwr_fwd *= 5;                                       // Full scale at 5W, 50W, 500W or 5kW.
            pwr_rev *= 5;
        }
        
        if (pwr_fwd != 0)                                       // Avoid dividing by 0.
        {
            a = (double)pwr_rev / (double)pwr_fwd;              // Calculate SWR ( r=sqrt(P_r/P_f) ; SWR=(1+r)/(1-r) ).
            a = sqrt(a);
            a = (1+a)/(1-a);
            swr = (int)(a*10);                                  // Multiply by 10 to have 0.1 resolution.
        }
        else                                                    // SWR is infinite.
            swr = -1;

        if ((pwr_fwd < 1000) && (pwr_rev < 1000))               // Switch LCD light on only when transmitting (power != 0).
        {
            LCD_LIGHT = LIGHT_OFF;
            if (SW_UNITS == 1)                                  // Display W or dBm?
            {
                LCD_home();                                     // Display idle message for W.
                LCD_line_1();
                LCD_write_string("P fwd          P ref");
                LCD_line_2();
                LCD_write_string("VSWR         <--VSWR");
            }
            else
            {
                LCD_home();                                     // Display idle message for dBm.
                LCD_line_1();
                LCD_write_string("Lp fwd        Lp ref");
                LCD_line_2();
                LCD_write_string("Return Loss  <--VSWR");
            }
        }
        else
        {
            LCD_LIGHT = LIGHT_ON;

            if (SW_UNITS == 1)                                  // Display W or dBm?
            {
                LCD_home();                                     // Display W.
                LCD_line_1();
                write_power(pwr_fwd / 10000);                   // Power is in the range 0..100, 0..200 or 0..500 (decimal point will be handled on the display).
                LCD_write_string("        ");
                write_power(pwr_rev / 10000);
                LCD_line_2();                                   // Display SWR.
                write_swr(swr);
    			LCD_write_string("    ");
                write_swr_bar(swr);
            }
            else
            {
                if (pwr_fwd != 0)                               // Avoid log(0) !!!
                {
                    a = 10 * log10(((double)pwr_fwd)/1000);     // Convert to dBm.
                    pwr_fwd = (int)a;
                    if (DP_1 == 0)                              // Handle decimal point.
                        pwr_fwd += 20;
                    if (DP_0 == 0)
                        pwr_fwd += 10;
                }
                else
                    pwr_fwd = -10000;                           // -infinite (-10000dBm is small enough).

                if (pwr_rev != 0)                               // Avoid log(0) !!!
                {
                    a = 10 * log10(((double)pwr_rev)/1000);     // Convert to dBm.
                    pwr_rev = (int)a;
                    if (DP_1 == 0)                              // Handle decimal point.
                        pwr_rev += 20;
                    if (DP_0 == 0)
                        pwr_rev += 10;
                }
                else
                    pwr_rev = -10000;                           // -infinite (-10000dBm is small enough).

                LCD_home();                                     // Display power in dBm.
                LCD_line_1();
                write_db(pwr_fwd);
                LCD_write_string("m      ");
                write_db(pwr_rev);
                LCD_write_string("m");
                LCD_line_2();                                   // Display return loss in dB and the usual SWR bar.
                write_db(pwr_fwd-pwr_rev);
                LCD_write_string("    ");
                write_swr_bar(swr);
            }
        }
    }
}


